import pandas as pd
from keras.layers import Multiply
from keras.layers.core import *
from keras.layers import Dense, Dropout, LSTM, Embedding, Activation, Lambda, Bidirectional, Flatten
from keras.engine import Input, Model, InputSpec
from keras.preprocessing.sequence import pad_sequences
import os
import copy
from keras.layers.normalization import BatchNormalization
import numpy as np
from sklearn.metrics import roc_auc_score
import random


# Hyperparameter settings
k_merList = [2]
kkk = k_merList[0]
BATCH_SIZE = 64 # a set of N samples. The samples in a batch are processed` independently,
                   #in parallel. If training, a batch results in only one update to the model.
kmer_words_size = 398

RNN_HIDDEN_DIM = 64    # Number of neural units in the lstm layer
DROPOUT_RATIO_1 = 0.5 # proportion of neurones not used for training
DROPOUT_RATIO_2 = 0.5 # proportion of neurones not used for training
xSquare_selection = 45 # Filter out the number of SeqPose Feature
maxlen = kmer_words_size - xSquare_selection    

checkpoint_dir = './checkpoints'  # Store model files
if not os.path.exists(checkpoint_dir):  # if the path does not exist, generate it
    os.mkdir(checkpoint_dir)

checkpoint_dir = './checkpoints/final_model'  # Store model files
if not os.path.exists(checkpoint_dir):  # if the path does not exist, generate it
    os.mkdir(checkpoint_dir)

useless_filename = './output/Useless_SeqPose_features_' + str(kkk) + '_mer_' + str(xSquare_selection) + '.csv'  # File storing sequence position
first_model_dir = checkpoint_dir + "/first_layer_model_"+str(kkk)+"_mer_"+str(xSquare_selection)+".hdf5"  # First layer model
second_model_dir = checkpoint_dir + "/second_layer_model_"+str(kkk)+"_mer_"+str(xSquare_selection)+".hdf5"  # Second layer model

def read_CSV():

    strong_test_dt = pd.read_csv('./data/strong-test-v1.csv', sep=',')
    weak_test_dt = pd.read_csv('./data/weak-test-v1.csv', sep=',')
    non_test_dt = pd.read_csv('./data/non-test-v1.csv', sep=',')
    independent_datasets_dt = strong_test_dt.append(weak_test_dt)
    independent_datasets_dt = independent_datasets_dt.append(non_test_dt)

    independent_datasets_dt = independent_datasets_dt.reset_index()
    independent_datasets_dt = independent_datasets_dt.drop(columns='index')

    independent_datasets_seqlist = independent_datasets_dt['sequence'].tolist()
    independent_datasets_tarlist = independent_datasets_dt['target'].tolist()
    independent_datasets_strengthlist = independent_datasets_dt['strength'].tolist()

    independent_datasets_list = []
    for i in range(len(independent_datasets_seqlist)):

        dict_tmp = {}
        dict_tmp['sequence'] = independent_datasets_seqlist[i]
        dict_tmp['target'] = independent_datasets_tarlist[i]
        dict_tmp['strength'] = independent_datasets_strengthlist[i]

        independent_datasets_list.append(dict_tmp)

    return independent_datasets_list

sequenceList = read_CSV()
random.seed(75)
random.shuffle(sequenceList)  # shuffle

# Use kmer to segment raw data
def k_split(sequenceList, k_merList):

    def k_mer(sequence, k):
        # kmer algorithm
        splitSequence = []  # Used to store the segmented sequence
        for i in range(len(sequence) - k + 1):
            tmp = ''
            for j in range(i, i + k):
                tmp += sequence[j]

            splitSequence.append(tmp)
        return splitSequence

    finalSequenceList = []
    for k_merNum in k_merList:  
        sequenceTmpList = copy.deepcopy(sequenceList) 
        for i in range(len(sequenceList)):  
            sequenceTmpList[i]['sequence'] = k_mer(sequenceList[i]['sequence'],
                                                   k_merNum)  
            finalSequenceList.append(sequenceTmpList[i])

    # Take each word as a feature,
    avertFinal = finalSequenceList[:len(sequenceList)]
    for i in range(len(k_merList)):
        tmplist = finalSequenceList[len(sequenceList) * i:(i + 1) * len(sequenceList)]
        for dictindex in range(len(tmplist)):
            avertFinal[dictindex]['sequence'] = avertFinal[dictindex]['sequence'] + tmplist[dictindex]['sequence']

    finalSequenceList = avertFinal 
    
    wordVectorSeqList = []  
    wordVectorTarList = [] 
    wordVectorStrengthList = []  

    for allDictSeq in finalSequenceList:
        wordVectorSeqList.append(allDictSeq['sequence'])  
        wordVectorTarList.append(allDictSeq['target']) 
        wordVectorStrengthList.append(allDictSeq['strength'])

    return wordVectorSeqList, wordVectorTarList, wordVectorStrengthList

wordVectorSeqList, wordVectorTarList, wordVectorStrengthList = k_split(sequenceList, k_merList)

# Load SeqPoseDict
token_dict = np.load('./output/SeqPoseDict.npy',allow_pickle=True).item()
SeqPoseDict_length = len(token_dict)

# Serialization
def get_index(sentence):
    global token_dict   
    sequence = []
    for word in sentence:
        try:
            sequence.append(token_dict[word])
        except KeyError:
            sequence.append(0)
    return sequence

def feature_selection(X_data, filename, xSquare_selection):  # Feature selection

    sorted_index_df = pd.read_csv(filename, header=None)
    sorted_index_list = sorted_index_df.iloc[:, 0].tolist()  

    unuseful_column = []
    for i in range(xSquare_selection):
        unuseful_column.append(sorted_index_list[i])
    unuseful_column.sort(reverse=True)  

    for i in range(len(X_data)): 
        for column_ele in unuseful_column:  
            del X_data[i][column_ele]

    return X_data

def isStandard(X_data_list):  

    fulfilled_list = []
    for index_list in X_data_list:
        if len(index_list) < kmer_words_size:
            cur_count = len(index_list)
            while cur_count < kmer_words_size:
                index_list.append(0)
                cur_count += 1
            fulfilled_list.append(index_list)

        elif len(index_list) > kmer_words_size:
            index_list_meet = index_list[:kmer_words_size]
            fulfilled_list.append(index_list_meet)

        elif len(index_list) == kmer_words_size:
            fulfilled_list.append(index_list)
    return fulfilled_list

X_data = list(map(get_index, wordVectorSeqList)) 
X_data = isStandard(X_data) 
# Feature selection
X_data = feature_selection(X_data, useless_filename, xSquare_selection)

# padding
maxlen_tmp = len(X_data[0]) 
if maxlen_tmp < maxlen:
    maxlen = maxlen_tmp  

# Partition data set
# build the training set and test set of the first layer
X_test_first = pad_sequences(X_data, maxlen=maxlen)  
Y_test_first = wordVectorTarList    

# build the training set and test set of the second layer
X_test_second = pad_sequences(X_data, maxlen=maxlen)   
Y_test_second = wordVectorStrengthList       

# Further build the training set required for the second layer model
final_X_test_second = []  
final_Y_test_second = []

for Seqindex in range(len(X_test_second)):  
  if Y_test_second[Seqindex]==1 or Y_test_second[Seqindex]==0:
    final_X_test_second.append(X_test_second[Seqindex])
    final_Y_test_second.append(Y_test_second[Seqindex])

# The processed second layer training set
final_X_test_second = pad_sequences(final_X_test_second, maxlen=maxlen)

# Model framework
def Model_TransBERT(input_length, SeqPoseDict_length):
    '''
    Define the classification model
    '''
    INPUT_DIM = SeqPoseDict_length + 1
    OUTPUT_DIM = 768
    
    def attention_3d_block(inputs, TIME_STEPS):     # attention
        input_dim = int(inputs.shape[2])
        a = Permute((2, 1))(inputs)
        a = Reshape((input_dim, TIME_STEPS))(a) # this line is not useful. It's just to know which dimension is what.
        a = Dense(TIME_STEPS, activation='softmax')(a)        
        a_probs = Permute((2, 1), name='attention_vec')(a)
        output_attention_mul = Multiply()([inputs, a_probs])
        return output_attention_mul
        
    def create_lstm(input_length, rnn_hidden_dim=RNN_HIDDEN_DIM, output_dim=OUTPUT_DIM, input_dim=INPUT_DIM,
                    dropout_1=DROPOUT_RATIO_1, dropout_2=DROPOUT_RATIO_2):
        
        inputs_1 = Input(shape=(input_length, ))      
        inputs_2 = Embedding(input_dim=INPUT_DIM, output_dim=output_dim, name='embedding_layer',
                  trainable=True)(inputs_1)
        
        bi_lstm_1 = Bidirectional(LSTM(rnn_hidden_dim, return_sequences=True))(inputs_2)
        bn = BatchNormalization()(bi_lstm_1)
        attention_mul_bilstm = attention_3d_block(bn, input_length)
        drop_1 = Dropout(DROPOUT_RATIO_1)(attention_mul_bilstm)
        bi_lstm_1 = Bidirectional(LSTM(rnn_hidden_dim))(drop_1)
        drop_2 = Dropout(DROPOUT_RATIO_2)(bi_lstm_1)

        dense = Dense(1, activation='sigmoid')(drop_2)
        
        model = Model(inputs_1, dense)
        
        model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
                
        return model
    
    return create_lstm(input_length)

def format_predict(y_pred_prob):
    y_pre = []
    for prob_ele in y_pred_prob:
        if prob_ele[0] >= 0.5: # class=1
            y_pre.append(1)
        else:
            y_pre.append(0)
    
    return y_pre

# Acc
def Model_ACC(model, X_test_first, Y_test_first, BATCH_SIZE):
    
    y_pred = model.predict(X_test_first, batch_size=BATCH_SIZE)
    y_pred = format_predict(y_pred)
    initial_positive = 0
    predict_right = 0  
    predict_wrong = 0  

    initial_negtive = 0 
    predict_negtive = 0  
    predict_right_negtive = 0  

    for i in range(len(y_pred)):
        predict_tmp = y_pred[i]
        initial_tmp = Y_test_first[i]  

        if int(initial_tmp) == 1:
            initial_positive += 1  
        if int(predict_tmp) == 1 and int(initial_tmp) == 1:
            predict_right += 1  
        if int(predict_tmp) == 0 and int(initial_tmp) == 1:
            predict_wrong += 1  

        if int(initial_tmp) == 0:
            initial_negtive += 1 
        if int(predict_tmp) == 1 and int(initial_tmp) == 0:
            predict_negtive += 1 
        if int(predict_tmp) == 0 and int(initial_tmp) == 0:
            predict_right_negtive += 1 

    if (predict_right + predict_negtive + predict_right_negtive + predict_wrong) != 0:
        res = (predict_right + predict_right_negtive) / (
                predict_right + predict_negtive + predict_right_negtive + predict_wrong)
    else:
        res = 0
    return res

# Sn
def Model_Sn(model, X_test_first, Y_test_first, BATCH_SIZE):

    y_pred = model.predict(X_test_first, batch_size=BATCH_SIZE)
    y_pred = format_predict(y_pred)
    initial_positive = 0
    predict_right = 0  
    predict_wrong = 0
    for i in range(len(y_pred)):
        predict_tmp = y_pred[i] 
        initial_tmp = Y_test_first[i]  

        if int(initial_tmp) == 1:
            initial_positive += 1 
        if int(predict_tmp) == 1 and int(initial_tmp) == 1:
            predict_right += 1  
        if int(predict_tmp) == 0 and int(initial_tmp) == 1:
            predict_wrong += 1  

    if (predict_right + predict_wrong) != 0:
        res = predict_right / (predict_right + predict_wrong)
    else:
        res = 0

    return res

# Sp
def Model_Sp(model, X_test_first, Y_test_first, BATCH_SIZE):
    
    y_pred = model.predict(X_test_first, batch_size=BATCH_SIZE)
    y_pred = format_predict(y_pred)

    initial_negtive = 0  
    predict_negtive = 0  
    predict_right_negtive = 0  

    for i in range(len(y_pred)):
        predict_tmp = y_pred[i]  
        initial_tmp = Y_test_first[i]  

        if int(initial_tmp) == 0:
            initial_negtive += 1 
        if int(predict_tmp) == 1 and int(initial_tmp) == 0:
            predict_negtive += 1  
        if int(predict_tmp) == 0 and int(initial_tmp) == 0:
            predict_right_negtive += 1 

    if (predict_right_negtive + predict_negtive) != 0:
        res = predict_right_negtive / (predict_right_negtive + predict_negtive)
    else:
        res = 0

    return res

# MCC
def Model_MCC(model, X_test_first, Y_test_first, BATCH_SIZE):
    
    y_pred = model.predict(X_test_first, batch_size=BATCH_SIZE)
    y_pred = format_predict(y_pred)
    initial_positive = 0
    predict_right = 0  
    predict_wrong = 0 

    initial_negtive = 0  
    predict_negtive = 0 
    predict_right_negtive = 0 

    for i in range(len(y_pred)):
        predict_tmp = y_pred[i]  
        initial_tmp = Y_test_first[i]  

        if int(initial_tmp) == 1:
            initial_positive += 1  
        if int(predict_tmp) == 1 and int(initial_tmp) == 1:
            predict_right += 1  
        if int(predict_tmp) == 0 and int(initial_tmp) == 1:
            predict_wrong += 1  

        if int(initial_tmp) == 0:
            initial_negtive += 1 
        if int(predict_tmp) == 1 and int(initial_tmp) == 0:
            predict_negtive += 1  
        if int(predict_tmp) == 0 and int(initial_tmp) == 0:
            predict_right_negtive += 1 

    fenzi = (predict_right * predict_right_negtive) - (predict_negtive * predict_wrong)
    fenmu = ((predict_right + predict_wrong) * (predict_right + predict_wrong) * (
                predict_right_negtive + predict_negtive) * (predict_right_negtive + predict_wrong)) ** 0.5
    if fenmu != 0:
        res = fenzi / fenmu
    else:
        res = 0

    return res

# AUC
def Model_AUC(model, X_test_first, Y_test_first, BATCH_SIZE):

    Test_y_pred_score = model.predict(X_test_first, batch_size=BATCH_SIZE)  

    Test_y_true = np.array(Y_test_first)
    Test_y_scores = np.array(Test_y_pred_score)
    return roc_auc_score(Test_y_true, Test_y_scores)

# load model
model_first = Model_TransBERT(len(X_test_first[0]), SeqPoseDict_length) 
model_first.load_weights(first_model_dir)

model_second = Model_TransBERT(len(final_X_test_second[0]), SeqPoseDict_length) 
model_second.load_weights(second_model_dir)

res = Model_ACC(model_first, X_test_first, Y_test_first, BATCH_SIZE)
print('The Acc of the first layer model on the test set is:', res)

res = Model_Sn(model_first, X_test_first, Y_test_first, BATCH_SIZE) 
print('The Sn of the first layer model on the test set is:', res)

res = Model_Sp(model_first, X_test_first, Y_test_first, BATCH_SIZE)
print('The Sp of the first layer model on the test set is:', res)

res = Model_MCC(model_first, X_test_first, Y_test_first, BATCH_SIZE)
print('The MCC of the first layer model on the test set is:', res)

res = Model_AUC(model_first, X_test_first, Y_test_first, BATCH_SIZE)  
print('The AUC of the first layer model on the test set is:', res)

res = Model_ACC(model_second, final_X_test_second, final_Y_test_second, BATCH_SIZE)
print('The Acc of the second layer model on the test set is:', res)

res = Model_Sn(model_second, final_X_test_second, final_Y_test_second, BATCH_SIZE)  
print('The Sn of the second layer model on the test set is:', res)

res = Model_Sp(model_second, final_X_test_second, final_Y_test_second, BATCH_SIZE)
print('The Sp of the second layer model on the test set is:', res)

res = Model_MCC(model_second, final_X_test_second, final_Y_test_second, BATCH_SIZE)
print('The MCC of the second layer model on the test set is:', res)

res = Model_AUC(model_second, final_X_test_second, final_Y_test_second, BATCH_SIZE)  
print('The AUC of the second layer model on the test set is:', res)
